/*
    qtZenOscillo : Software for using DSO-2020 usb scope under Linux
    Copyright (C) 2015 Olivier Poitevin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact : code@zener.fr

*/

#include "datathread.h"
#include <qmath.h>


// CMD :
#define CMD_CH1VOLT 0xe0
#define CMD_CH2VOLT 0xe1
#define CMD_SPEED   0xe2
#define CMD_TRIG    0xe3

#define BULKEP  6
#define MAXBULKDATASIZE 0x200000

// speed configurations :
typedef struct{
    int mant;
    int unit;
    unsigned char cmd;
    int nbSamples;
    int bulkDSize;
}Speed_t;


#define OSPEED(_mant_,_unit_,_cmd_,_nbSamples_,_bulkDSize_) {_mant_,_unit_,_cmd_,_nbSamples_,_bulkDSize_}

Speed_t speeds[]={
    OSPEED(1,UNIT_NS,0x30,131072,0x200000), // 48MHz WARNING
    OSPEED(2,UNIT_NS,0x30,131072,0x200000), // 48MHz WARNING
    OSPEED(5,UNIT_NS,0x30,131072,0x200000), // 48MHz WARNING
    OSPEED(10,UNIT_NS,0x30,131072,0x200000), // 48MHz WARNING
    OSPEED(20,UNIT_NS,0x30,131072,0x200000), // 48MHz WARNING
    OSPEED(50,UNIT_NS,0x30,24,0x200000), // 48MHz (48MHz * 10sq * 50*10e-9) = 24
    OSPEED(100,UNIT_NS,0x30,48,0x200000), // 48MHz
    OSPEED(200,UNIT_NS,0x30,96,0x200000), // 48MHz
    OSPEED(500,UNIT_NS,0x30,240,0x200000), // 48MHz
    OSPEED(1,UNIT_US,0x30,480,0x200000), // 48MHz
    OSPEED(2,UNIT_US,0x30,960,0x200000), // 48MHz (48MHz * 10sq * 2*10e-6) = 960
    OSPEED(5,UNIT_US,0x10,800,0x40000), // 16MHz (16MHz * 10sq * 5*10e-6) = 800
    OSPEED(10,UNIT_US,0x08,800,0x40000), // 8MHz (8MHz * 10sq * 10*10e-6) = 800
    OSPEED(20,UNIT_US,0x04,800,0x40000), // 4MHz (4MHz * 10sq * 20*10e-6) = 800
    OSPEED(50,UNIT_US,0x01,500,0x40000), // 1MHz
    OSPEED(100,UNIT_US,0x01,1000,0x40000), // 1MHz (1MHz * 10sq * 100*10e-6)=1000
    OSPEED(200,UNIT_US,0x01,2000,0x40000), // 1MHz
    OSPEED(500,UNIT_US,0x01,5000,0x40000), // 1MHz
    OSPEED(1,UNIT_MS,0x01,10000,0x40000), // 1MHz
    OSPEED(2,UNIT_MS,0x01,20000,0x40000), // 1MHz
    OSPEED(5,UNIT_MS,0x01,50000,0x100000), // 1MHz
    OSPEED(10,UNIT_MS,0x01,100000,0x100000), // 1MHz
    OSPEED(20,UNIT_MS,0x01,200000,0x100000), // 1MHz
    OSPEED(50,UNIT_MS,0x01,500000,0x200000), // 1MHz
    OSPEED(100,UNIT_MS,0x01,1000000,0x200000), // 1MHz
    OSPEED(200,UNIT_MS,0x32,1000000,0x200000), // 500kHz (500kHz * 10sq * 200*10e-3) = 1000000
    OSPEED(500,UNIT_MS,0x14,1000000,0x200000), // 200kHz
    OSPEED(1,UNIT_S,0x0a,131072,0x200000), // 100kHz

  {0,0,0,0,0}
};

DataThread::DataThread(QObject *parent) :
    QThread(parent)
{
    usbOpened=false;
    speedM=0;
    speedU=0;
    ch1Activated=false;
    ch2Activated=false;
    ch1Range=0;
    ch2Range=0;
    calibrated=false;
    timeNbSmp=0;
    bulkDataSize=0;
    fastMode=false;
}


DataThread::~DataThread()
{
    usbCom.closeDevice();
    usbOpened=false;
}

bool DataThread::trigNow()
{
    char dat[16];

    usbMtx.lock();
    dat[0]=0x01;
    if(!usbCom.usbWriteVendor(CMD_TRIG,0x00,0x00,dat,1,1000))
    {
        usbMtx.unlock();
        return false;
    }
    usbMtx.unlock();
    return true;
}

bool DataThread::initialize()
{
    stopThis=false;
    stopped=false;

    if(!usbCom.getUsbComDevice(DEFIDVENDOR,DEFIDPRODUCT))
    {
        // not found
        // set error string
        return false;
    }
    if(!usbCom.openDevice())
    {
        // cannot open
        // set error string
        return false;
    }
    usbOpened=true;

    if(!calibrated)
    {
        int speedUList[]={UNIT_MS,UNIT_US};
        int mVoltList[]={20,200,500,1000};
        // backup conf
        int ch1RangeBck=ch1Range;
        int ch2RangeBck=ch2Range;
        int speedMBck=speedM;
        int speedUBck=speedU;

        // calibrate now
        for(int speed=0;speed<2;speed++)
        {
            for(int mV=0;mV<4;mV++)
            {
                setSpeed(1,speedUList[speed]);
                setCh1Range(mVoltList[mV]);
                setCh2Range(mVoltList[mV]);
                if(!trigNow())
                    return false;
            }
        }
        // restore now
        ch1Range=ch1RangeBck;
        ch2Range=ch2RangeBck;
        speedM=speedMBck;
        speedU=speedUBck;

        calibrated=true;
    }

    // send configuration to device
    setSpeed(speedM,speedU);
    setCh1Range(ch1Range);
    setCh2Range(ch2Range);

    return true;
}

void DataThread::run()
{
    char dat[MAXBULKDATASIZE];

    while(true)
    {
        QVector<double> chan1Points;
        QVector<double> chan2Points;

        if(!trigNow())
        {
            sleep(1);
            continue;
        }

        bulkDataSizeMtx.lock();
        int curBulkDataSize=bulkDataSize;
        bulkDataSizeMtx.unlock();

        usbMtx.lock();
        if(!usbCom.usbBulkRead(BULKEP,dat,curBulkDataSize,2000))
        {
            sleep(1);
            continue;
        }
        usbMtx.unlock();

        for(int i=0;i<curBulkDataSize;i+=2)
        {
            // ch 1
            unsigned char v=*(unsigned char*)&dat[i];
            double v1;
            v1=(double)v-133.0;
            v1/=ch1Coef;
            if(v1>1.0) v1=1.0;
            if(v1<-1.0) v1=-1.0;
            chan1Points.append(v1);

            // ch2
            v=*(unsigned char*)&dat[i+1];
            double v2;
            v2=(double)v-133.0;
            v2/=ch2Coef;
            if(v2>1.0) v2=1.0;
            if(v2<-1.0) v2=-1.0;
            chan2Points.append(v2);
        }

        emit(dataReady(chan1Points,chan2Points,timeNbSmp));


        if(stopThis)
        {
            break;
        }
    }
    stopped=true;
}


void DataThread::doStop()
{
    stopThis=true;
    while(!this->stopped) msleep(100);
    usbCom.closeDevice();
    usbOpened=false;
}


bool DataThread::setSpeed(int mantissa,int unit)
{
    int i;

    speedM=mantissa;
    speedU=unit;
    timeNbSmp=0;

    if(!usbOpened)
    {
        return true;
    }

    i=0;
    while(speeds[i].mant)
    {
        if(mantissa==speeds[i].mant && unit==speeds[i].unit)
        {
            char dat[16];
            // found it
            timeNbSmp=speeds[i].nbSamples;
            dat[0]=speeds[i].cmd;

            bulkDataSizeMtx.lock();
            if(fastMode)
            {
                bulkDataSize=(timeNbSmp*2+0xfff)&~0xfff;
            }
            else
            {
                bulkDataSize=speeds[i].bulkDSize;
            }
            bulkDataSizeMtx.unlock();

            usbMtx.lock();
            if(!usbCom.usbWriteVendor(CMD_SPEED,0x00,0x00,dat,1,1000))
            {
                usbMtx.unlock();
                return false;
            }
            usbMtx.unlock();
            return true;
        }
        i++;
    }
    return false;
}

void DataThread::setCh1Activated(bool value)
{
    ch1Activated=value;
}

void DataThread::setCh2Activated(bool value)
{
    ch2Activated=value;
}

bool DataThread::setCh1Range(int mV)
{
    int lutmV[]={20,100,200,500,1000};
    int lutCmd[]={0x0a,0x0a,0x05,0x02,0x01};
    double chCoefs[]={5.0*10.0/2.0,5.0*50.0/2.0,5.0*50.0/2.0,5.0*24.5/1.0,5.0*25/1.0};

    // 20mV/sq : 10=>(2sq: 2/5) => 10/c=2/5 c=5*10/2
    // 100mV/sq : 50=>(2sq: 2/5) => 50/c=2/5 c=5*50/2
    // 200mV/sq : 50=>(2sq: 2/5) => 50/c=2/5 c=5*50/2
    // 500mV/sq : 24.5=>(1sq: 1/5) => 24.5/c=1/5 c=5*24.5/1
    // 1V/sq : 25=>(1sq: 1/5) => 25/c=1/5 c=5*25/1

    ch1Range=mV;
    ch1Coef=1.0;
    if(!usbOpened)
    {
        return true;
    }

    for(int i=0;i<5;i++)
    {
        if(lutmV[i]==mV)
        {
            char dat[16];

            ch1Coef=chCoefs[i];

            dat[0]=lutCmd[i];

            usbMtx.lock();
            if(!usbCom.usbWriteVendor(CMD_CH1VOLT,0x00,0x00,dat,1,1000))
            {
                usbMtx.unlock();
                return false;
            }
            usbMtx.unlock();
        }
    }
    return true;
}


bool DataThread::setCh2Range(int mV)
{
    int lutmV[]={20,100,200,500,1000};
    int lutCmd[]={0x0a,0x0a,0x05,0x02,0x01};
    double chCoefs[]={5.0*10.0/2.0,5.0*50.0/2.0,5.0*50.0/2.0,5.0*24.5/1.0,5.0*25/1.0};

    // 20mV/sq : 10=>(2sq: 2/5) => 10/c=2/5 c=5*10/2
    // 100mV/sq : 50=>(2sq: 2/5) => 50/c=2/5 c=5*50/2
    // 200mV/sq : 50=>(2sq: 2/5) => 50/c=2/5 c=5*50/2
    // 500mV/sq : 24=>(1sq: 1/5) => 24/c=1/5 c=5*24/1
    // 1V/sq : 25=>(1sq: 1/5) => 25/c=1/5 c=5*25/1

    ch2Range=mV;
    ch2Coef=1.0;

    if(!usbOpened)
    {
        return true;
    }

    for(int i=0;i<5;i++)
    {
        if(lutmV[i]==mV)
        {
            char dat[16];

            ch2Coef=chCoefs[i];

            dat[0]=lutCmd[i];

            usbMtx.lock();
            if(!usbCom.usbWriteVendor(CMD_CH2VOLT,0x00,0x00,dat,1,1000))
            {
                usbMtx.unlock();
                return false;
            }
            usbMtx.unlock();
        }
    }
    return true;
}

void DataThread::setFastMode(bool value)
{
    fastMode=value;
    setSpeed(speedM,speedU);
}


